/* dhcpc.c
 *
 * udhcp DHCP client
 *
 * Russ Dill <Russ.Dill@asu.edu> July 2001
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/file.h>
#include <unistd.h>
#include <getopt.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <time.h>
#include <string.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <errno.h>

#include "dhcpd.h"
#include "dhcpc.h"
#include "options.h"
#include "clientpacket.h"
#include "packet.h"
#include "script.h"
#include "socket.h"
#include "debug.h"
#include "pidfile.h"
#include "nvram.h"
#include "vendor.h"

static int state;
static unsigned long requested_ip; /* = 0 */
static unsigned long server_addr;
static unsigned long timeout;
static int packet_num; /* = 0 */
static int fd;
static int signal_pipe[2];
static int bootp_cnt_offer = 0/*, bootp_cnt_ack = 0*/;
#define LISTEN_NONE 0
#define LISTEN_KERNEL 1
#define LISTEN_RAW 2
static int listen_mode;

#define DEFAULT_SCRIPT	"/usr/share/udhcpc/default.script"

struct client_config_t client_config = {
	/* Default options. */
	abort_if_no_lease: 0,
	foreground: 0,
	quit_after_lease: 0,
	background_if_no_lease: 0,
	interface: "eth0",
	pidfile: NULL,
	script: DEFAULT_SCRIPT,
	clientid: NULL,
	hostname: NULL,
	ifindex: 0,
	arp: "\0\0\0\0\0\0",		/* appease gcc-3.0 */
};

#ifndef BB_VER
static void show_usage(void)
{
	printf(
"Usage: udhcpc [OPTIONS]\n\n"
"  -c, --clientid=CLIENTID         Client identifier\n"
"  -H, --hostname=HOSTNAME         Client hostname\n"
"  -h                              Alias for -H\n"
"  -f, --foreground                Do not fork after getting lease\n"
"  -b, --background                Fork to background if lease cannot be\n"
"                                  immediately negotiated.\n"
"  -i, --interface=INTERFACE       Interface to use (default: eth0)\n"
"  -n, --now                       Exit with failure if lease cannot be\n"
"                                  immediately negotiated.\n"
"  -p, --pidfile=file              Store process ID of daemon in file\n"
"  -q, --quit                      Quit after obtaining lease\n"
"  -r, --request=IP                IP address to request (default: none)\n"
"  -s, --script=file               Run file at dhcp events (default:\n"
"                                  " DEFAULT_SCRIPT ")\n"
"  -v, --version                   Display version\n"
	);
	exit(0);
}
#endif

char *fd2str(int fd)
{
	char *buf = NULL;
	size_t count = 0, n;

	do {
		buf = realloc(buf, count + 512);
		n = read(fd, buf + count, 512);
		if (n < 0) {
			xfree(buf);
			buf = NULL;
		}
		count += n;
	} while (n == 512);

	close(fd);
	if (buf)
		buf[count] = '\0';
	return buf;
}

char *file2str(const char *path)
{
	int fd;

	if ((fd = open(path, O_RDONLY)) == -1) {
		perror(path);
		return NULL;
	}

	return fd2str(fd);
}

/* just a little helper */
static void change_mode(int new_mode)
{
	dbg_printf("entering %s listen mode",
		new_mode ? (new_mode == 1 ? "kernel" : "raw") : "none");
	close(fd);
	fd = -1;
	listen_mode = new_mode;
}


/* perform a renew */
static void perform_renew(void)
{
	syslog( LOG_INFO,"Performing a DHCP renew");
	switch (state) {
	case BOUND:
		change_mode(LISTEN_KERNEL);
	case RENEWING:
	case REBINDING:
		state = RENEW_REQUESTED;
		break;
	case RENEW_REQUESTED: /* impatient are we? fine, square 1 */
		run_script(NULL, "deconfig");
	case REQUESTING:
	case RELEASED:
		change_mode(LISTEN_RAW);
		state = INIT_SELECTING;
		break;
	case INIT_SELECTING:
		break;
	}

	/* start things over */
	packet_num = 0;

	/* Kill any timeouts because the user wants this to hurry along */
	timeout = 0;
}


/* perform a release */
static void perform_release(void)
{
	char buffer[16];
	struct in_addr temp_addr;

	syslog( LOG_INFO,"perform_release start\n");
	/* send release packet */
	if (state == BOUND || state == RENEWING || state == REBINDING) {
		temp_addr.s_addr = server_addr;
		sprintf(buffer, "%s", inet_ntoa(temp_addr));
		temp_addr.s_addr = requested_ip;
		dbg_printf("Unicasting a release of %s to %s\n", 
				inet_ntoa(temp_addr), buffer);				
		send_release(server_addr, requested_ip); /* unicast */
		syslog( LOG_INFO,"sent release\n");
		run_script(NULL, "deconfig");
	}
	dbg_printf("Entering released state");

	change_mode(LISTEN_NONE);
	state = RELEASED;
	timeout = 0x7fffffff;
}


/* Exit and cleanup */
static void exit_client(int retval)
{
	char buf[10] = "";
	
	sprintf(buf, "%d", bootp_cnt_offer);
	nvram_set("bootp_cnt_offer", buf);
	//nvram_set("bootp_cnt_ack", bootp_cnt_ack);
	pidfile_delete(client_config.pidfile);
	CLOSE_LOG();
	exit(retval);
}


/* Signal handler */
static void signal_handler(int sig)
{
	if (write(signal_pipe[1], &sig, sizeof(sig)) < 0) {
		LOG(LOG_ERR, "Could not send signal: %s",
			strerror(errno));
	}
}


static void background(void)
{
	int pid_fd;

	pid_fd = pidfile_acquire(client_config.pidfile); /* hold lock during fork. */
	while (pid_fd >= 0 && pid_fd < 3) pid_fd = dup(pid_fd); /* don't let daemon close it */
	if (daemon(0, 0) == -1) {
		perror("fork");
		exit_client(1);
	}
	client_config.foreground = 1; /* Do not fork again. */
	pidfile_write_release(pid_fd);
}


#ifdef COMBINED_BINARY
int udhcpc_main(int argc, char *argv[])
#else
int main(int argc, char *argv[])
#endif
{
	unsigned char *temp, *message;
	unsigned long t1 = 0, t2 = 0, xid = 0;
	unsigned long start = 0, lease, ul_lease2;
	fd_set rfds;
	int retval;
	struct timeval tv;
	int c, len;
	struct dhcpMessage packet;
	struct in_addr temp_addr;
	int pid_fd;
	time_t now;
	int max_fd;
	int sig, time_interval;
	char *str;
	unsigned int up;
	int i=0;	
	u_int32_t val_u32;
	
	static struct option arg_options[] = {
		{"clientid",	required_argument,	0, 'c'},
		{"foreground",	no_argument,		0, 'f'},
		{"background",	no_argument,		0, 'b'},
		{"hostname",	required_argument,	0, 'H'},
		{"hostname",    required_argument,      0, 'h'},
		{"interface",	required_argument,	0, 'i'},
		{"now", 	no_argument,		0, 'n'},
		{"pidfile",	required_argument,	0, 'p'},
		{"quit",	no_argument,		0, 'q'},
		{"request",	required_argument,	0, 'r'},
		{"script",	required_argument,	0, 's'},
		{"version",	no_argument,		0, 'v'},
		{"help",	no_argument,		0, '?'},
		{0, 0, 0, 0}
	};

	/* get options */
	while (1) {
		int option_index = 0;
		c = getopt_long(argc, argv, "c:fbH:h:i:np:qr:s:v", arg_options, &option_index);
		if (c == -1) break;
		
		switch (c) {
		case 'c':
			len = strlen(optarg) > 255 ? 255 : strlen(optarg);
			if (client_config.clientid) xfree(client_config.clientid);
			client_config.clientid = xmalloc(len + 2);
			printf("dhcpc: line295, malloc=%d\n",len + 2);
			client_config.clientid[OPT_CODE] = DHCP_CLIENT_ID;
			client_config.clientid[OPT_LEN] = len;
			client_config.clientid[OPT_DATA] = '\0';
			strncpy(client_config.clientid + OPT_DATA, optarg, len);
			break;
		case 'f':
			client_config.foreground = 1;
			break;
		case 'b':
			client_config.background_if_no_lease = 1;
			break;
		case 'h':
		case 'H':
			len = strlen(optarg) > 255 ? 255 : strlen(optarg);
			if (client_config.hostname) xfree(client_config.hostname);
			client_config.hostname = xmalloc(len + 2);
			printf("dhcpc: line311, malloc=%d\n",len + 2);			
			client_config.hostname[OPT_CODE] = DHCP_HOST_NAME;
			client_config.hostname[OPT_LEN] = len;
			strncpy(client_config.hostname + 2, optarg, len);
			break;
		case 'i':
			client_config.interface =  optarg;
			break;
		case 'n':
			client_config.abort_if_no_lease = 1;
			break;
		case 'p':
			client_config.pidfile = optarg;
			break;
		case 'q':
			client_config.quit_after_lease = 1;
			break;
		case 'r':
			requested_ip = inet_addr(optarg);
			break;
		case 's':
			client_config.script = optarg;
			break;
		case 'v':
			syslog( LOG_INFO,"udhcpcd, version %s\n\n", VERSION);
			exit_client(0);
			break;
		default:
			show_usage();
		}
	}

	if(nvram_get("bootp_cnt_offer") == NULL)
	{
		bootp_cnt_offer = 0;
	}
	else
	{
		bootp_cnt_offer = atoi(nvram_get("bootp_cnt_offer"));
	}


//	if(nvram_get("bootp_cnt_ack") == NULL)
//	{
//		bootp_cnt_ack = 0;
//	}
//	else
//	{
//		bootp_cnt_ack = atoi(nvram_get("bootp_cnt_ack"));
//	}

	OPEN_LOG("udhcpc");
	syslog( LOG_INFO,"udhcp client (v%s) started", VERSION);

	pid_fd = pidfile_acquire(client_config.pidfile);
	pidfile_write_release(pid_fd);

	if (read_interface(client_config.interface, &client_config.ifindex, 
			   NULL, client_config.arp) < 0)
		exit_client(1);
		
	if (!client_config.clientid) {
		client_config.clientid = xmalloc(6 + 3);
			printf("dhcpc: line375, malloc=%d\n",6+3);		
		client_config.clientid[OPT_CODE] = DHCP_CLIENT_ID;
		client_config.clientid[OPT_LEN] = 7;
		client_config.clientid[OPT_DATA] = 1;
		memcpy(client_config.clientid + 3, client_config.arp, 6);
	}

	/* ensure that stdin/stdout/stderr are never returned by pipe() */
	if (fcntl(STDIN_FILENO, F_GETFL) == -1)
		(void) open("/dev/null", O_RDONLY);
	if (fcntl(STDOUT_FILENO, F_GETFL) == -1)
		(void) open("/dev/null", O_WRONLY);
	if (fcntl(STDERR_FILENO, F_GETFL) == -1)
		(void) open("/dev/null", O_WRONLY);	

	/* setup signal handlers */
	pipe(signal_pipe);
	signal(SIGUSR1, signal_handler);
	signal(SIGUSR2, signal_handler);
	signal(SIGTERM, signal_handler);
	
	state = INIT_SELECTING;
	run_script(NULL, "deconfig");
	change_mode(LISTEN_RAW);

	memset(&now,0,sizeof(now));
	for (;;) {
		dbg_printf("dhcpc:now1=%i\n",now);
		dbg_printf("dhcpc:timeout=%i\n",timeout);
		dbg_printf("dhcpc:timeout-now=%i\n",(timeout-now));	
		dbg_printf("dhcpc:state=%i\n",state);
		dbg_printf("dhcpc:packet_num=%i\n",packet_num);
				
		str = file2str("/proc/uptime");
		if (str) {
			up = atoi(str);
			xfree(str);
		}
		dbg_printf("dhcpc: uptime=%d\n",up);
//		tv.tv_sec = timeout - time(0);
		tv.tv_sec = timeout - up;
		dbg_printf("1 tv.tv_sec=%i\n",tv.tv_sec);		
		tv.tv_usec = 0;
		FD_ZERO(&rfds);
		dbg_printf("listen_socket\n");
		if (listen_mode != LISTEN_NONE && fd < 0) {
			if (listen_mode == LISTEN_KERNEL)
				fd = listen_socket(INADDR_ANY, CLIENT_PORT, client_config.interface);
			else
				fd = raw_socket(client_config.ifindex);
			if (fd < 0) {
				LOG(LOG_ERR, "FATAL: couldn't listen on socket, %s", strerror(errno));
				exit_client(0);
			}
		}
		if (fd >= 0) FD_SET(fd, &rfds);
		FD_SET(signal_pipe[0], &rfds);		
		dbg_printf("2 tv.tv_sec=%i\n",tv.tv_sec);
		if (tv.tv_sec > 0) {
			dbg_printf("Waiting on select...\n");
			max_fd = signal_pipe[0] > fd ? signal_pipe[0] : fd;
			retval = select(max_fd + 1, &rfds, NULL, NULL, &tv);
		} else retval = 0; /* If we already timed out, fall through */
//		now = time(0);
		str = file2str("/proc/uptime");
		if (str) {
			up = atoi(str);
			xfree(str);
		}
		now = up;
		dbg_printf("dhcpc:now2=%i\n",now);		
		dbg_printf("retval=%i\n",retval);								
		if (retval == 0) {
			/* timeout dropped to zero */
			switch (state) {
			case INIT_SELECTING:
// for dhcp test 2007.05.01
//				if (packet_num < 3) {
				if (packet_num==0){				
					if (packet_num == 0)
						xid = random_xid();

					/* send discover packet */
					send_discover(xid, requested_ip); /* broadcast */
					
					timeout = now + ((packet_num == 2) ? 4 : 2);
					packet_num++;
					time_interval++;
				} else {
#if 0					
					if (client_config.background_if_no_lease) {
						syslog( LOG_INFO,"No lease, forking to background.");
						background();
					} else if (client_config.abort_if_no_lease) {
						syslog( LOG_INFO,"No lease, failing.");
						exit_client(1);
				  	}
#endif				
					/* wait to try again */
#if 1					
					if((time_interval*20)<120)
						timeout = now + (time_interval*20);
					else
						timeout = now + 120;
					packet_num = 0 ;						
#else										
					packet_num = 0;
					timeout = now + 60;
#endif
				}
				break;
			case RENEW_REQUESTED:
			case REQUESTING:
				dbg_printf("Entering REQUESTING state, state=%i,packet_num=%i\n",state,packet_num);			
				if (packet_num < 3) {
					/* send request packet */
					if (state == RENEW_REQUESTED)
						send_renew(xid, server_addr, requested_ip); /* unicast */
					else send_selecting(xid, server_addr, requested_ip); /* broadcast */
					
					timeout = now + ((packet_num == 2) ? 10 : 2);
					packet_num++;
				} else {					
					/* timed out, go back to init state */
					dbg_printf("timed out, go back to init state and run run_script deconfig");						
					if (state == RENEW_REQUESTED) run_script(NULL, "deconfig");
					state = INIT_SELECTING;
					timeout = now;
					packet_num = 0;
					dbg_printf("change_mode to LISTEN_RAW");					
					change_mode(LISTEN_RAW);
				}
				break;
			case BOUND:
				/* Lease is starting to run out, time to enter renewing state */
				state = RENEWING;
				change_mode(LISTEN_KERNEL);
				dbg_printf("Entering BOUND state");
				/* fall right through */
			case RENEWING:
				dbg_printf("Entering RENEWING state\n");			
				/* Either set a new T1, or enter REBINDING state */
				if ((t2 - t1) <= (lease / 14400 + 1)) {
					dbg_printf( "(t2 - t1) <= (lease / 14400 + 1)");					
					/* timed out, enter rebinding state */
					state = REBINDING;
					timeout = now + (t2 - t1);
					dbg_printf("timed out,Entering rebinding state");
				} else {
					dbg_printf("(t2 - t1) > (lease / 14400 + 1) and send request packet\n");	
					if ((now-timeout)> 1000000000)
						dbg_printf("Got new ntp time, don't send request packets!!!");
					else
					{				
					/* send a request packet */
						send_renew(xid, server_addr, requested_ip); /* unicast */
					}
					t1 = (t2 - t1) / 2 + t1;
					timeout = t1 + start;
				}
				break;
			case REBINDING:
				/* Either set a new T2, or enter INIT state */
				if ((lease - t2) <= (lease / 14400 + 1)) {
					/* timed out, enter init state */
					state = INIT_SELECTING;
					dbg_printf("Lease lost, entering init state");
					run_script(NULL, "deconfig");
					timeout = now;
					packet_num = 0;
					change_mode(LISTEN_RAW);
				} else {
					/* send a request packet */
					dbg_printf("REBINDING and ((lease - t2) <= (lease / 14400 + 1))\n");
					dbg_printf("lease = %i,t2 =%i",lease, t2);										
					send_renew(xid, 0, requested_ip); /* broadcast */

					t2 = (lease - t2) / 2 + t2;
					timeout = t2 + start;
				}
				break;
			case RELEASED:
				/* yah, I know, *you* say it would never happen */
				timeout = 0x7fffffff;
				break;
			}
		} else if (retval > 0 && listen_mode != LISTEN_NONE && FD_ISSET(fd, &rfds)) {
			/* a packet is ready, read it */
			if (listen_mode == LISTEN_KERNEL)
				len = get_packet(&packet, fd);
			else len = get_raw_packet(&packet, fd);
			
			if (len == -1 && errno != EINTR) {
				syslog( LOG_INFO,"error on read, %s, reopening socket", strerror(errno));
				change_mode(listen_mode); /* just close and reopen */
			}
			if (len < 0) continue;
			
			if (packet.xid != xid) {
				syslog( LOG_INFO,"Ignoring XID %lx (our xid is %lx)",
					(unsigned long) packet.xid, xid);
				continue;
			}
			
			if ((message = get_option(&packet, DHCP_MESSAGE_TYPE)) == NULL) {
				DEBUG(LOG_ERR, "couldnt get option from packet -- ignoring");
				continue;
			}
			
			switch (state) {
			case INIT_SELECTING:
				/* Must be a DHCPOFFER to one of our xid's */
				if (*message == DHCPOFFER) {
					
					//dbg_printf("udhcpc.c[%d] : [%d] [%d]\n", __LINE__, packet.sname[0], packet.file[0]);
					
					if(!memcmp(packet.chaddr, client_config.arp,6))
					{
						syslog( LOG_INFO,"Get DHCP OFFER packet!!!\n");
					}
					else
					{
						syslog( LOG_INFO,"Get DHCP OFFER packet --- not for us!!!\n");
						break;
					}
					
					if((strlen(packet.sname) != 0) || (strlen(packet.file) != 0))
					{
/*
						if( packet.file[0] < 128 )
						{
								dbg_printf("Get BOOTP data...ignore\n");
								break;
						}		
*/
						bootp_cnt_offer++;	

						if(bootp_cnt_offer >= 2)
						{
							syslog( LOG_INFO,"DHCP Offer : Get BOOTP data for %d times. Pass!!!\n", bootp_cnt_offer);
						}
						else
						{
							syslog( LOG_INFO,"DHCP Offer : Get BOOTP data for %d times. Ignore!!!\n", bootp_cnt_offer);
							timeout += 2;
							break;
						}
					}
					
					bootp_cnt_offer = 0;
								
					if ((temp = get_option(&packet, DHCP_SERVER_ID))) {
						memcpy(&server_addr, temp, 4);
						xid = packet.xid;
						requested_ip = packet.yiaddr;
						
						/* enter requesting state */
						state = REQUESTING;
						timeout = now;
						packet_num = 0;
					} else {
						DEBUG(LOG_ERR, "No server ID in message");
					}
				}
				break;
			case RENEW_REQUESTED:
			case REQUESTING:
			case RENEWING:
			case REBINDING:
				if (*message == DHCPACK) {

					//dbg_printf("udhcpc.c[%d] : [%d] [d]\n", __LINE__, packet.sname[0], packet.file[0]);
					
					if(!memcmp(packet.chaddr, client_config.arp,6))
					{
						syslog( LOG_INFO,"Get DHCP ACK packet!!!\n");
					}
					else
					{
						syslog( LOG_INFO,"Get DHCP ACK packet --- not for us!!!\n");
						break;
					}					
					
//					if((strlen(packet.sname) != 0) || (strlen(packet.file) != 0))
//					{
///*						
//						if( packet.file[0] < 128 )
//						{
//								dbg_printf("Get BOOTP data...ignore\n");
//								break;
//						}		
//*/
//
//						bootp_cnt_ack++;	
//						
//						if(bootp_cnt_ack == 3)
//						{
//							dbg_printf("DHCP ACK : Get BOOTP data for 3 times. Pass!!!\n");
//						}
//						else
//						{
//							dbg_printf("DHCP ACK : Get BOOTP data for %d times. Ignore!!!\n", bootp_cnt_ack);
//							timeout += 2;
//							break;
//						}						
//					}
//					
//					bootp_cnt_ack = 0;
						
					temp = get_option(&packet, (unsigned char)DHCP_LEASE_TIME_CODE);											
					dbg_printf("temp lease time=%s\n",temp);	
					if (temp==NULL) {
						syslog( LOG_INFO,"No lease time with ACK, using 1 hour lease");
						lease = 60 * 60;
					} else {						
						memcpy(&lease, temp, 4);						
						lease = ntohl(lease);
						dbg_printf("lease=%i\n",lease);						
					}
						
					/* enter bound state */
					dbg_printf("enter bound state\n");					
					t1 = lease / 2;
					
					/* little fixed point for n * .875 */
					t2 = (lease * 0x7) >> 3;
					dbg_printf("t1=%i, t2=%i\n",t1,t2);						
					temp_addr.s_addr = packet.yiaddr;
					dbg_printf("Lease of %s obtained, lease time %ld\n", 
						inet_ntoa(temp_addr), lease);
						
					start = now;
					timeout = t1 + start;
					requested_ip = packet.yiaddr;
					dbg_printf("run_script with state=%i\n",state);
					dbg_printf("run_script\n");					
					run_script(&packet,((state == RENEWING || state == REBINDING) ? "renew" : "bound"));

					state = BOUND;
					dbg_printf("change_mode to LISTEN_NONE\n");					
					change_mode(LISTEN_NONE);
					dbg_printf("quit_after_lease=%i,foreground=%i\n",client_config.quit_after_lease,client_config.foreground);

					if (client_config.quit_after_lease) 
						exit_client(0);
/*					if (!client_config.foreground)
						background();
*/						

				} else if (*message == DHCPNAK) {
					/* return to init state */
					syslog( LOG_INFO,"Received DHCP NAK");
					run_script(&packet, "nak");
					if (state != REQUESTING)
						run_script(NULL, "deconfig");
					state = INIT_SELECTING;
					timeout = now;
					requested_ip = 0;
					packet_num = 0;
					change_mode(LISTEN_RAW);
					sleep(3); /* avoid excessive network traffic */
				}
				break;
			/* case BOUND, RELEASED: - ignore all packets */
			}	
		} else if (retval > 0 && FD_ISSET(signal_pipe[0], &rfds)) {
			if (read(signal_pipe[0], &sig, sizeof(sig)) < 0) {
				syslog( LOG_INFO,"Could not read signal: %s", 
					strerror(errno));
				continue; /* probably just EINTR */
			}
			dbg_printf("sig=%i\n");			
			switch (sig) {
			case SIGUSR1: 
				perform_renew();
				break;
			case SIGUSR2:
				perform_release();
				break;
			case SIGTERM:
				syslog( LOG_INFO,"Received SIGTERM\n");
				exit_client(0);
			}
		} else if (retval == -1 && errno == EINTR) {
			/* a signal was caught */		
		} else {
			/* An error occured */
			syslog( LOG_INFO,"Error on select");
		}
		
	}
	return 0;
}

